#ifndef _OPENDRIVE_LANE_H
#define _OPENDRIVE_LANE_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_road_node.h"
#include "opendrive_road_segment.h"
#include "opendrive_lane.h"

class COpendriveLane
{
  friend class COpendriveRoadSegment;
  friend class COpendriveRoadNode;
  friend class COpendriveRoad;
  private:
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<COpendriveLane *> next;
    std::vector<COpendriveLane *> prev;
    COpendriveLane *left_lane;
    opendrive_mark_t left_mark;
    COpendriveLane *right_lane;
    opendrive_mark_t right_mark;
    COpendriveRoadSegment *segment;
    double scale_factor;
    double resolution;
    double width;
    double speed;
    double offset;
    int id;
    unsigned int index;
  protected:
    COpendriveLane();
    COpendriveLane(const COpendriveLane &object);
    void load(const ::lane &lane_info,COpendriveRoadSegment *segment);
    void link_neightbour_lane(COpendriveLane *lane);
    void link_lane(COpendriveLane *lane,opendrive_mark_t mark,bool from_start,bool to_start,bool belongs_to_lane);
    void add_next_lane(COpendriveLane *lane);
    void add_prev_lane(COpendriveLane *lane);
    void remove_lane(COpendriveLane *lane);
    void add_node(COpendriveRoadNode *node);
    bool has_node(COpendriveRoadNode *node);
    bool has_node_with_index(unsigned int index);
    void set_parent_segment(COpendriveRoadSegment *segment);
    void set_left_lane(COpendriveLane *lane,opendrive_mark_t mark);
    void set_right_lane(COpendriveLane *lane,opendrive_mark_t mark);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_width(double width);
    void set_speed(double speed);
    void set_offset(double offset);
    void set_id(int id);
    void set_index(unsigned int index);
    bool updated(lane_up_ref_t &refs);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void update_start_node(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *start=NULL);
    void update_end_node(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *end=NULL);
    COpendriveLane *get_sub_lane(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *start=NULL,TOpendriveWorldPose *end=NULL);
    COpendriveLane *clone(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref);
  public:
    unsigned int get_num_nodes(void) const;
    const COpendriveRoadNode &get_node(unsigned int index) const;
    const COpendriveRoadSegment &get_segment(void) const;
    unsigned int get_num_next_lanes(void) const;
    const COpendriveLane &get_next_lane(unsigned int index) const;
    unsigned int get_num_prev_lanes(void) const;
    const COpendriveLane &get_prev_lane(unsigned int index) const;
    double get_width(void) const;
    double get_speed(void) const;
    double get_offset(void) const;
    double get_center_offset(void) const;
    unsigned int get_index(void) const;
    int get_id(void) const;
    TOpendriveWorldPose get_start_pose(void) const;
    TOpendriveWorldPose get_end_pose(void) const;
    double get_length(void) const;
    TOpendriveWorldPose get_pose_at(double length);
    double get_curvature_at(double length);    
    TOpendriveLocalPose transform_to_local_pose(TOpendriveTrackPose &track) const;
    TOpendriveWorldPose transform_to_world_pose(TOpendriveTrackPose &track) const;
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    friend std::ostream& operator<<(std::ostream& out, COpendriveLane &lane);
    ~COpendriveLane();
};

#endif
