#ifndef _OPENDRIVE_ROAD_H
#define _OPENDRIVE_ROAD_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_road_segment.h"
#include "opendrive_road_node.h"
#include "opendrive_lane.h"

class COpendriveRoad
{
  friend class COpendriveRoadSegment;
  friend class COpendriveRoadNode;
  friend class COpendriveLane;
  private:
    std::vector<COpendriveRoadSegment *> segments;
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<COpendriveLane *> lanes;
    double resolution;
    double scale_factor;
    double min_road_length;
  protected:
    COpendriveRoadSegment *get_segment_by_id(std::string &id);
    void free(void);
    void link_segments(OpenDRIVE &open_drive);
    unsigned int add_node(COpendriveRoadNode *node);
    bool remove_node(COpendriveRoadNode *node);
    COpendriveRoadNode* get_node_at(TOpendriveWorldPose &pose);
    bool node_exists_at(TOpendriveWorldPose &pose);
    unsigned int add_lane(COpendriveLane *lane);
    bool remove_lane(COpendriveLane *lane);
    void complete_open_lanes(void);
    void add_segment(COpendriveRoadSegment *segment);
    bool has_segment(COpendriveRoadSegment *segment);
    std::vector<unsigned int> update_path(node_up_ref_t &node_refs,std::vector<unsigned int> &path);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void reindex(void);
    void prune(std::vector<unsigned int> &path_nodes);
  public:
    COpendriveRoad();
    COpendriveRoad(const COpendriveRoad& object);
    void load(const std::string &filename);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_min_road_length(double length);
    double get_resolution(void) const;
    double get_scale_factor(void) const;
    double get_min_road_length(void) const;
    unsigned int get_num_segments(void) const;
    const COpendriveRoadSegment& get_segment(unsigned int index) const;
    unsigned int get_num_lanes(void) const;
    const COpendriveLane &get_lane(unsigned int index) const;
    unsigned int get_num_nodes(void) const;
    const COpendriveRoadNode& get_node(unsigned int index) const;
    const COpendriveRoadSegment &operator[](std::size_t index);
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    const COpendriveLane &get_closest_lane(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    unsigned int get_closest_lane_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    const COpendriveRoadSegment &get_closest_segment(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    unsigned int get_closest_segment_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    void get_closest_poses(TOpendriveWorldPose &pose,std::vector<TOpendriveWorldPose> &closest_poses,std::vector<double> &closest_lengths,double dist_tol,double angle_tol=0.1) const;
    std::vector<unsigned int> get_sub_road(std::vector<unsigned int> &path_nodes,TOpendriveWorldPose &start_pose,TOpendriveWorldPose &end_pose,COpendriveRoad &new_road);
    void get_sub_road(TOpendriveWorldPose &start_pose,double &length,double margin,COpendriveRoad &new_road);
    void operator=(const COpendriveRoad& object);
    friend std::ostream& operator<<(std::ostream& out, COpendriveRoad &road);
    ~COpendriveRoad();
};

#endif
