#ifndef _OPENDRIVE_ROAD_NODE_H
#define _OPENDRIVE_ROAD_NODE_H

#include <vector>
#include "opendrive_common.h"
#include "opendrive_link.h"
#include "opendrive_lane.h"
#include "opendrive_road_segment.h"

typedef struct
{
  COpendriveLane * lane;
  COpendriveRoadSegment *segment;
  double start_curvature;
  double end_curvature;
}TOpendriveRoadNodeParent;

class COpendriveRoadNode
{
  friend class COpendriveLane;
  friend class COpendriveRoadSegment;
  friend class COpendriveRoad;
  friend class COpendriveLink;
  private:
    std::vector<COpendriveLink *> links;
    double resolution;
    double scale_factor;
    TOpendriveWorldPose pose;
    std::vector<TOpendriveRoadNodeParent> parents;
    unsigned int index;
  protected:
    COpendriveRoadNode();
    COpendriveRoadNode *clone(link_up_ref_t &new_link_ref);
    void free(void);
    bool add_link(COpendriveRoadNode *node,opendrive_mark_t mark,COpendriveRoadSegment *segment,COpendriveLane *lane);
    void add_link(COpendriveLink *link);
    void remove_link(COpendriveLink *link);
    bool has_link(COpendriveLink *link);
    COpendriveLink *get_link_with(COpendriveRoadNode *node);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_index(unsigned int index);
    void set_pose(TOpendriveWorldPose &start);
    void add_parent(COpendriveLane *lane,COpendriveRoadSegment *segment,double start_curvature,double end_curvature);
    TOpendriveRoadNodeParent *get_parent_with_lane(const COpendriveLane *lane);
    TOpendriveRoadNodeParent *get_parent_with_segment(const COpendriveRoadSegment *segment);
    bool updated(node_up_ref_t &refs);
    COpendriveRoadNode *get_original_node(node_up_ref_t &node_refs);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void update_start_pose(COpendriveLane *lane,TOpendriveWorldPose &start,double distance);
    void update_end_pose(COpendriveLane *lane,TOpendriveWorldPose &end,double distance);
  public:
    double get_resolution(void) const;
    double get_scale_factor(void) const;
    unsigned int get_index(void) const;
    unsigned int get_num_parents(void) const;
    const COpendriveRoadSegment& get_parent_segment(unsigned int index) const;
    const COpendriveLane &get_parent_lane(unsigned int index) const;
    TOpendriveWorldPose get_pose(void) const;
    unsigned int get_num_links(void) const;
    const COpendriveLink &get_link(unsigned int index) const;
    const COpendriveLink &get_link_ending_at(unsigned int node_index) const;
    bool has_link_with(const COpendriveRoadNode &node) const;
    unsigned int get_closest_link(TOpendriveWorldPose &pose,double angle_tol=0.1) const;
    double get_closest_distance(TOpendriveWorldPose &pose,bool only_lanes=false,double angle_tol=0.1) const;
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,bool only_lanes=false,double angle_tol=0.1) const;
    void get_closest_poses(TOpendriveWorldPose &pose,std::vector<TOpendriveWorldPose> &closest_poses,std::vector<double> &closest_lengths,double dist_tol,bool only_lanes=false,double angle_tol=0.1) const;
    friend std::ostream& operator<<(std::ostream& out, COpendriveRoadNode &node);
    ~COpendriveRoadNode();
};

#endif
